#include "stdafx.h"
#include "RunProcess.h"

/* Make linker happy without CTR */

extern "C"
{
#pragma function(memset)
	void *memset(void *dest, int c, size_t count)
	{
		char *bytes = (char *)dest;
		while (count--)
		{
			*bytes++ = (char)c;
		}
		return dest;
	}

#pragma function(memcpy)
	void *memcpy(void *dest, const void *src, size_t count)
	{
		char *dest8 = (char *)dest;
		const char *src8 = (const char *)src;
		while (count--)
		{
			*dest8++ = *src8++;
		}
		return dest;
	}

}

//source: ReactOS' itow.c
wchar_t * CDECL _ultow(unsigned long value, wchar_t *string, int radix)
{
	WCHAR buffer[33];
	PWCHAR pos;
	WCHAR digit;

	pos = &buffer[32];
	*pos = '\0';

	do {
		digit = (WCHAR)(value % radix);
		value = value / radix;
		if (digit < 10) {
			*--pos = '0' + digit;
		}
		else {
			*--pos = 'a' + digit - 10;
		} /* if */
	} while (value != 0L);

	if (string != NULL) {
		memcpy(string, pos, (&buffer[32] - pos + 1) * sizeof(WCHAR));
	} /* if */
	return string;
}

/*
Some WOW64 applications won't correctly load wow64log.dll if it has imports from Kernel32
(because it will try to load it and fail with STATUS_CONFLICTING_ADDRESSES), so we have
to stick to NtDll... yay.
*/

#ifdef LOG_TO_FILE

void Log(wchar_t * str, NTSTATUS lErrorCode) {
	HANDLE hFile = NULL;
	OBJECT_ATTRIBUTES objAttribs = { 0 };
	size_t dwBytesToWrite = 0;
	StringCbLengthW(str, 4096, &dwBytesToWrite);
	DWORD dwBytesWritten = 0;

	PCWSTR filePath = L"\\??\\C:\\Temp\\W64L.log";
	UNICODE_STRING unicodeString;
	RtlInitUnicodeString(&unicodeString, filePath);

	InitializeObjectAttributes(&objAttribs, &unicodeString, OBJ_CASE_INSENSITIVE, NULL, NULL);

	const int allocSize = 2048;
	LARGE_INTEGER largeInteger;
	largeInteger.QuadPart = allocSize;

	IO_STATUS_BLOCK ioStatusBlock = { 0 };

	NtCreateFile(&hFile, FILE_APPEND_DATA | SYNCHRONIZE, &objAttribs, &ioStatusBlock, &largeInteger,
		FILE_ATTRIBUTE_NORMAL, 0, FILE_OPEN_IF, FILE_SEQUENTIAL_ONLY | FILE_SYNCHRONOUS_IO_NONALERT, NULL, 0);

	if (hFile == INVALID_HANDLE_VALUE || hFile == NULL)
		return;

	NtWriteFile(hFile, NULL, NULL, NULL, &ioStatusBlock, str, dwBytesToWrite, NULL, NULL);

	if (lErrorCode) {
		wchar_t sErrorCode[33];
		_ultow(lErrorCode, sErrorCode, 16);
		StringCbLengthW(sErrorCode, 33, &dwBytesToWrite);
		NtWriteFile(hFile, NULL, NULL, NULL, &ioStatusBlock, L"\t", 2, NULL, NULL);
		NtWriteFile(hFile, NULL, NULL, NULL, &ioStatusBlock, sErrorCode, dwBytesToWrite, NULL, NULL);
	}

	NtWriteFile(hFile, NULL, NULL, NULL, &ioStatusBlock, L"\r\n", 4, NULL, NULL);

	NtClose(hFile);
}

#else

// no Kernel32, no DebugOutputString
void Log(wchar_t * str, NTSTATUS lErrorCode) {}

#endif

/* The following function won't properly run all executables, but it's good enough to run CMD */

void __stdcall RunProcess(PWSTR Path, PWSTR Cmd)
{
	NTSTATUS status;
	UNICODE_STRING ImagePathName;
	UNICODE_STRING CommandLine;
	PRTL_USER_PROCESS_PARAMETERS ProcessParams;
	RTL_USER_PROCESS_INFORMATION ProcessInfo;

	if (!RtlDosPathNameToNtPathName_U(Path, &ImagePathName, NULL, NULL))
	{
		Log(L"RtlDosPathNameToNtPathName_U Path failed", 0);
		return;
	}

	RtlInitUnicodeString(&CommandLine, Cmd);

	status = RtlCreateProcessParameters(&ProcessParams, &ImagePathName, NULL, NULL, &CommandLine, NULL, NULL, NULL, NULL, NULL);
	if (status)
	{
		Log(L"RtlCreateProcessParameters failed", status);
		return;
	}

	status = RtlCreateUserProcess(&ImagePathName, OBJ_CASE_INSENSITIVE, ProcessParams, NULL, NULL, NULL, FALSE, NULL, NULL, &ProcessInfo);
	if (status)
	{
		Log(L"RtlCreateUserProcess failed", status);
		RtlDestroyProcessParameters(ProcessParams);
		return;
	}

	NtResumeThread(ProcessInfo.ThreadHandle, NULL);

	RtlDestroyProcessParameters(ProcessParams);
	NtClose(ProcessInfo.ThreadHandle);
	NtClose(ProcessInfo.ProcessHandle);

	Log(L"Everything OK", 0);
	return;
}

